package dev.shadowsoffire.apotheosis.adventure.affix;

import dev.shadowsoffire.apotheosis.Apotheosis;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.StringTag;
import dev.shadowsoffire.placebo.util.StepFunction;
import java.util.Collection;
import net.minecraft.world.entity.projectile.AbstractArrow;
import net.minecraft.world.entity.Entity;
import dev.shadowsoffire.placebo.codec.CodecProvider;
import java.util.stream.Stream;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityRegistry;
import java.util.function.Function;
import dev.shadowsoffire.placebo.util.CachedObject;
import java.util.Collections;
import javax.annotation.Nullable;
import net.minecraft.network.chat.Component;
import java.util.Iterator;
import net.minecraft.nbt.Tag;
import net.minecraft.nbt.CompoundTag;
import java.util.Map;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import java.util.HashMap;
import net.minecraft.world.item.ItemStack;
import net.minecraft.resources.ResourceLocation;

public class AffixHelper
{
    public static final ResourceLocation AFFIX_CACHED_OBJECT;
    public static final String DISPLAY = "display";
    public static final String LORE = "Lore";
    public static final String AFFIX_DATA = "affix_data";
    public static final String AFFIXES = "affixes";
    public static final String RARITY = "rarity";
    public static final String NAME = "name";
    public static final String CATEGORY = "category";
    
    public static void applyAffix(final ItemStack stack, final AffixInstance affix) {
        final HashMap<DynamicHolder<? extends Affix>, AffixInstance> affixes = new HashMap<DynamicHolder<? extends Affix>, AffixInstance>(getAffixes(stack));
        affixes.put(affix.affix(), affix);
        setAffixes(stack, affixes);
    }
    
    public static void setAffixes(final ItemStack stack, final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes) {
        final CompoundTag afxData = stack.m_41698_("affix_data");
        final CompoundTag affixesTag = new CompoundTag();
        for (final AffixInstance inst : affixes.values()) {
            affixesTag.m_128350_(inst.affix().getId().toString(), inst.level());
        }
        afxData.m_128365_("affixes", (Tag)affixesTag);
    }
    
    public static void setName(final ItemStack stack, final Component name) {
        final CompoundTag afxData = stack.m_41698_("affix_data");
        afxData.m_128359_("name", Component.Serializer.m_130703_(name));
    }
    
    @Nullable
    public static Component getName(final ItemStack stack) {
        if (!stack.m_41782_()) {
            return null;
        }
        final CompoundTag afxData = stack.m_41737_("affix_data");
        if (afxData == null) {
            return null;
        }
        return (Component)Component.Serializer.m_130701_(afxData.m_128461_("name"));
    }
    
    public static Map<DynamicHolder<? extends Affix>, AffixInstance> getAffixes(final ItemStack stack) {
        if (AffixRegistry.INSTANCE.getValues().isEmpty()) {
            return Collections.emptyMap();
        }
        return (Map)CachedObject.CachedObjectSource.getOrCreate(stack, AffixHelper.AFFIX_CACHED_OBJECT, (Function)AffixHelper::getAffixesImpl, CachedObject.hashSubkey("affix_data"));
    }
    
    public static Map<DynamicHolder<? extends Affix>, AffixInstance> getAffixesImpl(final ItemStack stack) {
        if (stack.m_41619_()) {
            return Collections.emptyMap();
        }
        final Map<DynamicHolder<? extends Affix>, AffixInstance> map = new HashMap<DynamicHolder<? extends Affix>, AffixInstance>();
        final CompoundTag afxData = stack.m_41737_("affix_data");
        if (afxData != null && afxData.m_128441_("affixes")) {
            final CompoundTag affixes = afxData.m_128469_("affixes");
            DynamicHolder<LootRarity> rarity = getRarity(afxData);
            if (!rarity.isBound()) {
                rarity = (DynamicHolder<LootRarity>)RarityRegistry.getMinRarity();
            }
            final LootCategory cat = LootCategory.forItem(stack);
            for (final String key : affixes.m_128431_()) {
                final DynamicHolder<Affix> affix = (DynamicHolder<Affix>)AffixRegistry.INSTANCE.holder(new ResourceLocation(key));
                if (affix.isBound()) {
                    if (!((Affix)affix.get()).canApplyTo(stack, cat, (LootRarity)rarity.get())) {
                        continue;
                    }
                    final float lvl = affixes.m_128457_(key);
                    map.put(affix, new AffixInstance(affix, stack, rarity, lvl));
                }
            }
        }
        return Collections.unmodifiableMap((Map<? extends DynamicHolder<? extends Affix>, ? extends AffixInstance>)map);
    }
    
    public static Stream<AffixInstance> streamAffixes(final ItemStack stack) {
        return getAffixes(stack).values().stream().filter(AffixInstance::isValid);
    }
    
    public static boolean hasAffixes(final ItemStack stack) {
        return stack.m_41782_() && !stack.m_41783_().m_128469_("affix_data").m_128469_("affixes").m_128456_();
    }
    
    public static void setRarity(final ItemStack stack, final LootRarity rarity) {
        final CompoundTag afxData = stack.m_41698_("affix_data");
        afxData.m_128359_("rarity", RarityRegistry.INSTANCE.getKey((CodecProvider)rarity).toString());
    }
    
    public static void copyFrom(final ItemStack stack, final Entity entity) {
        if (stack.m_41782_() && stack.m_41737_("affix_data") != null) {
            final CompoundTag afxData = stack.m_41737_("affix_data").m_6426_();
            afxData.m_128359_("category", LootCategory.forItem(stack).getName());
            entity.getPersistentData().m_128365_("affix_data", (Tag)afxData);
        }
    }
    
    @Nullable
    public static LootCategory getShooterCategory(final Entity entity) {
        final CompoundTag afxData = entity.getPersistentData().m_128469_("affix_data");
        if (afxData != null && afxData.m_128441_("category")) {
            return LootCategory.byId(afxData.m_128461_("category"));
        }
        return null;
    }
    
    public static Map<DynamicHolder<? extends Affix>, AffixInstance> getAffixes(final AbstractArrow arrow) {
        final Map<DynamicHolder<? extends Affix>, AffixInstance> map = new HashMap<DynamicHolder<? extends Affix>, AffixInstance>();
        final CompoundTag afxData = arrow.getPersistentData().m_128469_("affix_data");
        if (afxData != null && afxData.m_128441_("affixes")) {
            final CompoundTag affixes = afxData.m_128469_("affixes");
            DynamicHolder<LootRarity> rarity = getRarity(afxData);
            if (!rarity.isBound()) {
                rarity = (DynamicHolder<LootRarity>)RarityRegistry.getMinRarity();
            }
            for (final String key : affixes.m_128431_()) {
                final DynamicHolder<Affix> affix = (DynamicHolder<Affix>)AffixRegistry.INSTANCE.holder(new ResourceLocation(key));
                if (!affix.isBound()) {
                    continue;
                }
                final float lvl = affixes.m_128457_(key);
                map.put(affix, new AffixInstance(affix, ItemStack.f_41583_, rarity, lvl));
            }
        }
        return map;
    }
    
    public static Stream<AffixInstance> streamAffixes(final AbstractArrow arrow) {
        return getAffixes(arrow).values().stream();
    }
    
    public static DynamicHolder<LootRarity> getRarity(final ItemStack stack) {
        if (!stack.m_41782_()) {
            return (DynamicHolder<LootRarity>)RarityRegistry.INSTANCE.emptyHolder();
        }
        final CompoundTag afxData = stack.m_41737_("affix_data");
        return getRarity(afxData);
    }
    
    public static DynamicHolder<LootRarity> getRarity(@Nullable final CompoundTag afxData) {
        if (afxData != null) {
            try {
                return (DynamicHolder<LootRarity>)RarityRegistry.byLegacyId(afxData.m_128461_("rarity"));
            }
            catch (final IllegalArgumentException e) {
                afxData.m_128473_("rarity");
                return (DynamicHolder<LootRarity>)RarityRegistry.byLegacyId("empty");
            }
        }
        return (DynamicHolder<LootRarity>)RarityRegistry.INSTANCE.emptyHolder();
    }
    
    public static Collection<DynamicHolder<Affix>> byType(final AffixType type) {
        return AffixRegistry.INSTANCE.getTypeMap().get((Object)type);
    }
    
    public static StepFunction step(final float min, final int steps, final float step) {
        return new StepFunction(min, steps, step);
    }
    
    public static void addLore(final ItemStack stack, final Component lore) {
        final CompoundTag display = stack.m_41698_("display");
        final ListTag tag = display.m_128437_("Lore", 8);
        tag.add((Object)StringTag.m_129297_(Component.Serializer.m_130703_(lore)));
        display.m_128365_("Lore", (Tag)tag);
    }
    
    static {
        AFFIX_CACHED_OBJECT = Apotheosis.loc("affixes");
    }
}
